
# ------- CUSTOM PERL EXTENSION --------

package XML::JHXML;

use vars qw($VERSION);

$VERSION = "0.1.8";

use constant ELEMENT_NODE    => 1;
use constant ATTRIBUTE_NODE  => 2;
use constant TEXT_NODE	     => 3;
use constant DOCUMENT_NODE   => 9;
use constant NYI => "#NOT_YET_IMPLEMENTED";


sub new {
  my $class = ref($_[0]) || $_[0];
  $opts = NewParserOptions();
  return bless [$opts], $class;
}

sub recovering {
  my ($self,$value)=@_;
  return unless ref($self->[0]);
  if (@_>1) {
    $value = $value ? 1 : 0;
    SetOptionParseErrorsOK($self->[0],$value);
    return $value;
  } else {
    return ParserOptions_fParseErrorsOK_get($self->[0]);
  }
}

sub keep_blanks {
  my ($self,$value)=@_;
  return unless ref($self->[0]);
  if (@_>1) {
    if ($value == -1) {
      SetOptionNoBlanks($self->[0],2);
    } elsif ($value == 0) {
      SetOptionNoBlanks($self->[0],1);
    } elsif ($value == 1) {
      SetOptionNoBlanks($self->[0],0);
    } else {
      warn "JHXML: value $value for keep_blanks not supported\n";
    }
    return $value;
  } else {
    return ParserOptions_fNoBlanks($self->[0]);
  }
}

sub indent {
  my ($self,$value)=@_;
  return unless ref($self->[0]);
  if (@_>1) {
    SetOptionIndent($self->[0],$value-1);
    return $value;
  } else {
    return ParserOptions_iIndent($self->[0]);
  }
}

sub DESTROY {
  my ($self)=@_;
  DeleteParserOptions($self->[0]);
}

sub parse_file   {
  my ($self,$file)=@_;
  ParseFile($file,$self->[0]);
}

sub parse_string {
  ParseString($_[1], $_[0]->[0])
}

package XML::JHXML::Document;

sub nodeType { XML::JHXML->DOCUMENT_NODE };
sub getNodeType { XML::JHXML->DOCUMENT_NODE };
sub nodeName { "" };
sub toFile {
  XML::JHXML::WriteFile(@_);
}


*documentElement      = \&XML::JHXML::DocGetRootElement;
*getDocumentElement   = \&XML::JHXML::DocGetRootElement;
*createElement	      = \&XML::JHXML::CreateElement;
*createTextNode	      = \&XML::JHXML::CreateTextNode;

*destroy	      = \&XML::JHXML::DestroyDoc;

*_docSetAttribute     = \&XML::JHXML::DocSetAttribute;
*_docRemoveAttribute  = \&XML::JHXML::DocRemoveAttribute;
*_docAppendText	      = \&XML::JHXML::DocAppendText;

package XML::JHXML::Attr;

sub nodeType { XML::JHXML->ATTRIBUTE_NODE };
sub getNodeType { XML::JHXML->ATTRIBUTE_NODE };

# TO BE REIMPLEMENTED BY NATIVE LIBRARY WRAPPERS
*nodeName = \&XML::JHXML::Attr_pszName_get;
*value    = \&XML::JHXML::Attr_pszValue_get;
*setValue = \&XML::JHXML::Attr_pszValue_set;
*next     = \&XML::JHXML::GetNextAttr;

*getName = *nodeName;
*name = *nodeName;
*getValue = *value;

sub toString { 
  my ($self)=@_;
  $self->nodeName().'="'.$self->value().'" '
}


package XML::JHXML::Node;

sub nodeType {
  return $_[0]->isTextNode ? XML::JHXML->TEXT_NODE : XML::JHXML->ELEMENT_NODE;
}

*getNodeType	  = *nodeType;

*toString	  = \&XML::JHXML::ToString;
*parentNode	  = \&XML::JHXML::GetParent;
*getParentNode	  = \&XML::JHXML::GetParent;
*firstChild	  = \&XML::JHXML::GetFirstChild;
*getFirstChild	  = \&XML::JHXML::GetFirstChild;
*lastChild	  = \&XML::JHXML::GetLastChild;
*getLastChild	  = \&XML::JHXML::GetLastChild;
*nextSibling	  = \&XML::JHXML::GetNextElement;
*getNextSibling	  = \&XML::JHXML::GetNextElement;
*previousSibling    = \&XML::JHXML::GetPrevElement;
*getPreviousSibling = \&XML::JHXML::GetPrevElement;
*firstAttribute	  = \&XML::JHXML::GetFirstAttr;
*getAttribute	  = \&XML::JHXML::GetAttribute;
*setAttribute	  = \&XML::JHXML::SetAttribute;
*getData	  = \&XML::JHXML::GetData;

*nodeName	  = \&XML::JHXML::GetElementName;
*getNodeName	  = \&XML::JHXML::GetElementName;
*localname	  = \&XML::JHXML::GetElementName;
*getLocalName	  = \&XML::JHXML::GetElementName;
*isSameNode	  = \&XML::JHXML::IsSameNode;
*hasChildNodes	  = \&XML::JHXML::HasChildNodes;
*isTextNode	  = \&XML::JHXML::IsTextNode;

*insertBefore	  = \&XML::JHXML::InsertBefore;
*insertAfter	  = \&XML::JHXML::InsertAfter;
*appendChild	  = \&XML::JHXML::AppendChild;
*removeChild	  = \&XML::JHXML::RemoveChild;
*replaceChild	  = \&XML::JHXML::ReplaceChild;

*destroyThisNode  = \&XML::JHXML::DestroyNode;
*destroy	  = \&XML::JHXML::DestroySubtree;

*removeAttribute  = \&XML::JHXML::RemoveAttribute;
*setData	  = \&XML::JHXML::SetData;

*appendText	  = \&XML::JHXML::AppendText;
*appendTextChild  = \&XML::JHXML::AppendTextChild;

*findFirstChild   = \&XML::JHXML::FindFirstChild;
*findAncestor     = \&XML::JHXML::FindAncestor;
*findNextSibling  = \&XML::JHXML::FindNextSibling;
*findPreviousSibling  = \&XML::JHXML::FindPrevSibling;

sub getChildrenByTagName {
  my ($self,$name)=@_;
  if (wantarray) {
    my @ret;
    my $node=$self->findFirstChild($name);
    while ($node) {
      push @ret,$node;
      $node = $node->findNextSibling($name);
    }
    return @ret;
  } else {
    # optimized version
    my $ret=0;
    my $node=$self->findFirstChild($name);
    while ($node) {
      $ret++;
      $node = $node->findNextSibling($name);
    }
    return $ret;
  }
}

sub getElementsByTagName {
  my ($self,$name)=@_;
  if (wantarray) {
    my @ret;
    my $node=$self->firstChild();
    while ($node) {
      push @ret,$node
	if ($node->nodeType == XML::JHXML->ELEMENT_NODE
	    and ($name eq '*' or $node->nodeName() eq $name));
      $node = $node->followingNode($self);
    }
    return @ret;
  } else {
    # optimized version
    my $ret=0;
    my $node=$self->firstChild();
    while ($node) {
      $ret++ if ($node->nodeType == XML::JHXML->ELEMENT_NODE
		 and ($name eq '*'  or $node->nodeName() eq $name));
      $node = $node->followingNode($self);
    }
    return $ret;
  }
}

# convenient extension (not in DOM)
sub followingNode {
  my ($node,$top)=@_;
  if ($node->firstChild) {
    return $node->firstChild;
  }
  do {
    return $node->nextSibling if $node->nextSibling;
    $node = $node->parentNode;
    return undef if ($node and $node->isSameNode($top));
  } while ($node);
  return undef;
}

# convenient extension (not in DOM)
sub precedingNode {
  my ($node,$top)=@_;
  if ($node->previousSibling) {
    $node=$node->previousSibling;
    while ($node->hasChildNodes) {
      $node=$node->lastChild;
    }
    return $node;
  } elsif ($node->parentNode
	   and
	   !$node->parentNode->isSame($top)) {
    return $node->parentNode;
  } else {
    return undef;
  }
}

package XML::JHXML;

1;

# ------- SWIG GENERATED SHADOW CLASSES --------


