#!/usr/bin/perl

package TrEd::EngValLex::PropBankEditor;

use strict;
use Tk;
use Tk::BrowseEntry;
use Tk::Label;
use Tk::Adjuster;
use Tk::ROText;
use File::Basename;

use TrEd::EngValLex::LibXMLPropBank;

my $ProbBankPath;
my $XMLDataClass;

my $doc;

#--- Widget design ---------------------------------------------
# frames
my ($fr_roleset, $fr_args, $fr_examples, $label_link);
my $label_predicate;
my $optionmenu_rolesets;
my ($scrolled_args, $scrolled_examples, $label_rolesetname);
my $filename;

my @bold   = (-foreground=>'#C00000');
my @selected = (-background => '#C9D8E3', qw/-relief raised -borderwidth 1/);
my @normal = (-background => undef, qw/-relief flat/);

#---------------------------------------------------------------


# Reference to the EngValLex editor if the PropBankViewer is integrated in it
my $editor;

# Selected roleset
my $roleset;

# List of rolesets
my $rolesets;

# Last selected link (hash with fields type and reference)
my %link = (type=>0, text=>0);  

# GUI ----------------------------------------------------------------------
sub setLink {
    $link{type} = shift;
    $link{rolesetref} = shift;
    $link{fileref} = basename($filename);
    
    if ($link{type} eq 'PB roleset') {
	$link{text} = $link{type}."[fileref: ".$link{fileref}.", rolesetref: ".$link{rolesetref}."]";
    } elsif ($link{type} eq 'PB example') {
	$link{name} = shift;
	$link{annotation} = shift;
	$link{annotation} =~ s/^ *//g;
	$link{annotation} =~ s/  / /g;
	$link{text} = $link{type}."[fileref: ".$link{fileref}.", rolesetref: ".$link{rolesetref}.", name: ".$link{name}."]";
    } elsif ($link{type} eq 'PB role') {
	$link{n} = shift;
	$link{f} = shift;
	$link{text} = $link{type}."[fileref: ".$link{fileref}.", rolesetref: ".$link{rolesetref}.", n: ".$link{n}.", f: ".$link{f}."]";
    }
    
    $label_link->configure(-text=>"LINK: ".$link{text});

    if (defined($editor)) {
	$editor->setPropBankLink(\%link);
    }
}


sub addArgs {
    $scrolled_args->delete(0.1, "end");
    my @roles = TrEd::EngValLex::LibXMLPropBank->get_roles($doc, shift);
    for my $role (@roles) {
        my $n = $role->{n};
	my $f = $role->{f};
        my $descr = $role->{descr};
        $scrolled_args->insert('end', "Arg ${n}: ${descr}", 'arg'.${n});
        $scrolled_args->insert('end', "\n");
	$scrolled_args->tagBind('arg'.${n}, '<Any-Enter>' =>
            sub {shift->tagConfigure('arg'.${n}, @selected)}
        );
	$scrolled_args->tagBind('arg'.${n}, '<Any-Leave>' =>
            sub {shift->tagConfigure('arg'.${n}, @normal)}
        );
        $scrolled_args->tagBind('arg'.${n}, '<Button-1>' =>
            sub { setLink("PB role", $roleset, ${n}, ${f}); }
        );
    }
}

sub addExamples {
    $scrolled_examples->delete(0.1, "end");
    my @examples = TrEd::EngValLex::LibXMLPropBank->get_examples($doc, shift);
    my $i=0;
    for my $example (@examples) {
        my $name = $example->{name};
        my $text = $example->{text};
        my $label = 'example'.${i};
        $scrolled_examples->insert('end', $name.": ", 'jmeno');
        $scrolled_examples->insert('end', $text, $label);
        $scrolled_examples->insert('end', "\n");
	$scrolled_examples->tagBind($label, '<Any-Enter>' =>
            sub {shift->tagConfigure($label, @selected)}
        );
	$scrolled_examples->tagBind($label, '<Any-Leave>' =>
            sub {shift->tagConfigure($label, @normal)}
        );
        $scrolled_examples->tagBind($label, '<Button-1>' =>
            sub { setLink("PB example", $roleset, $name, $text); }
        );
        $i++;
    }
    $scrolled_examples->tagConfigure('jmeno', @bold);
}

sub setPredicate {
    $label_predicate->configure(-text=>shift) if $label_predicate;
}

sub setRoleSets {
    $rolesets = shift;
    
    my @opts = map{$_->{id}} @{$rolesets};
    $optionmenu_rolesets->configure(-options=>[@opts]);
    $roleset = $rolesets->[0]->{'id'};
}

sub change_predicate {
    my $top = shift;
    my @types =  (["XML files", ['.xml']]);
    my $filename = $top->getOpenFile(-filetypes => \@types, -initialdir => $ProbBankPath);
    if (defined $filename and $filename ne '') {
        print "opening $filename ...\n";
        loadPBFile($filename);
    }
}

sub changeRoleset {
    my ($self, $rs) = @_;
    $roleset = $rs;
    print STDERR "roleset changed to $rs\n";
    addArgs($rs);
    addExamples($rs);
    return unless defined($rolesets);
    my $n_roleset = (grep {$_->{id} eq $rs} @{$rolesets})[0];
    return unless defined($n_roleset);
    my $name = $n_roleset->{'name'};
    $label_rolesetname->configure(-text => $name);
    setLink("PB roleset", $roleset);
}

sub checkArgLink {
    my ($self, $roleset, $n, $f) = @_;
    return 1 if !defined ($doc);
    return TrEd::EngValLex::LibXMLPropBank->checkArgLink($doc, $roleset, $n, $f);
}

# create the PropBank widget, fills in default values
sub createPBWidget {
    $editor = shift;
    my $top = shift;
    if (! ref($top)) { $top = shift; }
    my $fe_confs = shift;
    if ($fe_confs && $fe_confs->{PropBank_dir}) {
	$ProbBankPath = $fe_confs->{PropBank_dir};
    }
    
    $fr_roleset = $top->Frame()->pack(-side=>'top', -fill=>'x', -expand=>'1');
    my $adj1 = $top->Adjuster(-widget=>$fr_roleset, -side=>'top')->pack(-fill=>'x');
    $fr_args = $top->Frame()->pack(-side=>'top',
				   -fill=>'both',
				   -expand=>'1');
    my $adj2 = $top->Adjuster(-widget=>$fr_args, -side=>'top')->pack(-fill=>'x');
    $fr_examples = $top->Frame()->pack(-side=>'top', -fill=>'both', -expand=>'1');
    
    # fr_roleset
    $label_link = $fr_roleset->Label(-text=>'link:',
				  -foreground=>'#00a',
				  -justify=>'left')
	->pack(-side=>'bottom', -fill=>'y');
    $label_rolesetname = $fr_roleset->Label()->pack(-side=>'bottom', -fill=>'x');
    $fr_roleset->Label(-text => 'Predicate: ') -> pack(-side=>'left', -fill=>'y');
    $label_predicate = $fr_roleset->Label(-text => '') -> pack(-side=>'left', -fill=>'y');
    $fr_roleset->Button(-text => '...', -command => sub { change_predicate($top); })
	->pack(-side=>'left');   
    $optionmenu_rolesets = $fr_roleset->Optionmenu(-options => [()],
                                                   -variable => \$roleset,
                                                   -command => sub { changeRoleset(0,shift); })
	->pack(-side=>'left');

    
    # fr_args
    $scrolled_args = $fr_args->Scrolled('ROText', -scrollbars=>'osoe', -wrap=>'word', -height=>"10")->pack(-fill=>'x', -expand=>'1');
    addArgs();
    
    # fr_examples
    $scrolled_examples = $fr_examples->Scrolled('ROText', -scrollbars=>'osoe', -wrap=>'word', -height=>"10")->pack(-fill=>'x', -expand=>'1');
    addExamples();
    

    my $roleset_frame = $top->Frame->pack;
    my $roles_frame = $top->Frame->pack;
    my $examples_frame = $top->Frame->pack;
    
}


# DATA ---------------------------------------------------------------------
# loads data from a single PropBank file and fills the PBWidget
sub loadPBFile {
    $filename = shift;
# if (not (-e $filename) ) {  
#        print STDERR "File $filename not found.\n";
#        return;
#	}
 
	eval { require TrEd::EngValLex::LibXMLPropBank; };
    die $@ if $@;
    my $parser;
    eval{
		($parser, $doc) = TrEd::EngValLex::LibXMLPropBank->parser_start($filename, 1);
	};

	return if $@;
	
    setPredicate(TrEd::EngValLex::LibXMLPropBank->get_predicate($doc));
    my @rolesets = TrEd::EngValLex::LibXMLPropBank->get_rolesets($doc);
    setRoleSets(\@rolesets);
}

# loads data from a single PropBank file and fills the PBWidget.
# Tries to gues the filename from the lemma
sub loadPBLemma {
    my $filename = shift;
    if (! ref($filename)) { $filename = shift; }
    loadPBFile($ProbBankPath.(($^O eq "MSWin32") ? "\\" : "/").$filename.".xml");
}


# to be called when run stand-alone
sub start {
    my $top = MainWindow->new(-title => 'PropBank viewer');
    $top->geometry('400x400');
    createPBWidget(undef, $top);
    loadPBFile($ProbBankPath . (($^O eq "MSWin32") ? "\\" : "/") . "force.xml");
    MainLoop;
}

#start();
1;
